/*
 *  chardev.c: Creates a read-only char device that says how many times
 *  you've read from the dev file
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>	/* for put_user */
#include <linux/rwsem.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/uaccess.h>
#include <linux/sched.h>

#define DEVICE_NAME              "chardev"	/* Dev name as it appears in /proc/devices   */
#define SUCCESS                  0
#define BUF_LEN                  80	    			/* Max length of the message from the device */

/**************************************  Quantitative      ***************************************/
#define _H_K			         0x200
#define _K				         (2 * _H_K)
#define _M				         (_K * _K)
#define _G				         (_M * _M)

#define MEM_CHUNK_SIZE 		     PAGE_SIZE

/************************************** SysLog Definitions ***************************************/
#define __FILE_NAME__ 			(strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)
#define LOG_STD_SEPARATOR 		"|"
#define LOG_STD_PREFIX 			LOG_STD_SEPARATOR"%s"LOG_STD_SEPARATOR"%s"LOG_STD_SEPARATOR"%04d"LOG_STD_SEPARATOR /*|File Name|Function|Line Number|*/

/***************************************  Log Entry Macros ***************************************/
#define klogprintln(log_prefix, fmt, ...)\
								printk(log_prefix""LOG_STD_PREFIX""fmt "\n",\
								__FILE_NAME__, __FUNCTION__, __LINE__, ##__VA_ARGS__)

#define LOGLEVEL_DEFAULT        -1          /* default (or last) loglevel */
#define LOGLEVEL_EMERG          0           /* system is unusable */
#define LOGLEVEL_ALERT          1           /* action must be taken immediately */
#define LOGLEVEL_CRIT           2           /* critical conditions */
#define LOGLEVEL_ERR            3           /* error conditions */
#define LOGLEVEL_WARNING        4           /* warning conditions */
#define LOGLEVEL_NOTICE         5           /* normal but significant condition */
#define LOGLEVEL_INFO           6           /* informational */
#define LOGLEVEL_DEBUG          7           /* debug-level messages */

#define LOGLBL_DEFAULT          "DFT"       /*-1 default (or last) loglevel */
#define LOGLBL_EMERG            "EMERGENCY" /*0  system is unusable */
#define LOGLBL_ALERT            "ALERT"     /*1  action must be taken immediately */
#define LOGLBL_CRIT             "CRTICAL"   /*2  critical conditions */
#define LOGLBL_ERR              "ERROR"     /*3  error conditions */
#define LOGLBL_WARNING          "WARNING"   /*4  warning conditions */
#define LOGLBL_NOTICE           "NOTICE"    /*5  normal but significant condition */
#define LOGLBL_INFO             "INFO"      /*6  informational */
#define LOGLBL_DEBUG            "DBG"       /*7  debug-level messages */

#define CATLOGDEFS(postfix)								KERN_##postfix""LOG_STD_SEPARATOR""LOGLBL_##postfix
#define CATLOGLEVEL(postfix)							LOGLEVEL_##postfix
#define LOGLEVEL_COND_EXPR(postfix)						if(CATLOGLEVEL(postfix) <= debug_level)
#define CAT_LOG_ENTRY_BLOCK(postfix, fmt, ...) 			LOGLEVEL_COND_EXPR(postfix)\
			klogprintln (CATLOGDEFS(postfix), fmt, ##__VA_ARGS__)

#define kalertprintln(fmt, ...) CAT_LOG_ENTRY_BLOCK(ALERT  , fmt, ##__VA_ARGS__)
#define kcritprintln(fmt, ...)  CAT_LOG_ENTRY_BLOCK(CRIT   , fmt, ##__VA_ARGS__)
#define kerrprintln(fmt, ...)   CAT_LOG_ENTRY_BLOCK(ERR    , fmt, ##__VA_ARGS__)
#define kwrnprintln(fmt, ...)   CAT_LOG_ENTRY_BLOCK(WARNING, fmt, ##__VA_ARGS__)
#define knoteprintln(fmt, ...)  CAT_LOG_ENTRY_BLOCK(NOTICE , fmt, ##__VA_ARGS__)
#define kinfoprintln(fmt, ...)  CAT_LOG_ENTRY_BLOCK(INFO   , fmt, ##__VA_ARGS__)
#define kdbgprintln(fmt, ...)   CAT_LOG_ENTRY_BLOCK(DEBUG  , fmt, ##__VA_ARGS__)

unsigned int mem_pool_size = 100 * _K;
unsigned int debug_level = LOGLEVEL_DEBUG;

module_param(mem_pool_size, int, S_IRUGO | S_IWUSR );
/* debug level parameter */
module_param(debug_level, int, S_IRUGO | S_IWUSR );


/*
 *  Prototypes - this would normally go in a .h file
 */
int 				init_module				(void);
void 				cleanup_module			(void);

static int 			device_open				(struct inode *, struct file *);
static int 			device_release			(struct inode *, struct file *);
static ssize_t 		device_read				(struct file *, char *, size_t, loff_t *);
static ssize_t 		device_write			(struct file *, const char *, size_t, loff_t *);

static int			chdv_mmap				(struct file *filp, struct vm_area_struct *vma);


/*
 * Global variables are declared as static, so are global within the file.
 */
static int Major;		/* Major number assigned to our device driver */
static int _dev_open = 0;	/* Is device open?
				 * Used to prevent multiple access to device */
static char msg[BUF_LEN];	/* The msg the device will give when asked */
static char *msg_Ptr;
static struct rw_semaphore rw_sem;
static void * g_membuff  = NULL;

static struct file_operations fops = {
	.owner              = THIS_MODULE,
	.mmap               = chdv_mmap,
	.read 				= device_read,
	.write 				= device_write,
	.open 				= device_open,
	.release 			= device_release
};
/*
 * This function is called when the module is loaded
 */
int init_module(void)
{
	int rc = SUCCESS;

	Major = register_chrdev(0, DEVICE_NAME, &fops);

	if (Major >= 0)
	{
		init_rwsem(&rw_sem);
		mem_pool_size = ALIGN(mem_pool_size, PAGE_SIZE);

		kdbgprintln("I was assigned major number %d. To talk to", Major);
		printk( "\t\tthe driver, create a dev file with\n");
		printk( "\t\t'mknod /dev/%s c %d 0'.\n", DEVICE_NAME, Major);
		printk( "\t\tTry various minor numbers. Try to cat and echo to\n");
		printk( "\t\tthe device file.\n");
		printk( "\t\tRemove the device file and module when done.\n");
	}
	else
	{
		kalertprintln("Registering char device failed with %d", Major);
		rc =  Major;
	}
	return rc;
}
/*
 * This function is called when the module is unloaded
 */
void cleanup_module(void)
{

	if(NULL != g_membuff)
	{
		kdbgprintln("Releasing memory...");
		kfree(g_membuff);
		g_membuff = NULL;
	}

//	if(NULL != m_class)
//	{
//		kdbgprintln("Destroying device class...");
//		class_destroy(m_class);
//		m_class = NULL;
//	}

	kdbgprintln("Unregister the device...");
	unregister_chrdev(Major, DEVICE_NAME);
}
/*
 * Methods
 */

/*
 * Called when a process tries to open the device file, like
 * "cat /dev/mycharfile"
 */
static int device_open(struct inode *inode, struct file *file)
{
    if( !down_read_trylock(&rw_sem) )
    {
    	kerrprintln("Failed to obtain read lock");
        return -EBUSY;
    }
    try_module_get(THIS_MODULE);
    sprintf(msg, "I already told you %d times Hello world!", _dev_open);
    _dev_open++;
	msg_Ptr = msg;

	return SUCCESS;
}

/*
 * Called when a process closes the device file.
 */
static int device_release(struct inode *inode, struct file *file)
{

//	kdbgprintln( "Dev ref cnt %d", _dev_open);
//	_dev_open--;		/* We're now ready for our next caller */
//
//	if(0 == _dev_open)
//	{
//		/*
//		 * Decrement the usage count, or else once you opened the file, you'll
//		 * never get get rid of the module.
//		 */
//		kdbgprintln( "Releasing the device");
//		module_put(THIS_MODULE);
//	}

//	kdbgprintln( "Dev ref cnt %d, private data ptr %p", _dev_open, file->private_data);
	_dev_open--;
//    kfree(file->private_data);
    up_read(&rw_sem);
    module_put(THIS_MODULE);

	return 0;
}

/*
 * Called when a process, which already opened the dev file, attempts to
 * read from it.
 */
static ssize_t device_read(struct file *filp,	/* see include/linux/fs.h   */
			   char *buffer,	/* buffer to fill with data */
			   size_t length,	/* length of the buffer     */
			   loff_t * offset)
{
	/*
	 * Number of bytes actually written to the buffer
	 */
	int bytes_read = 0;
	char * msg_idx = msg_Ptr;

	/*
	 * If we're at the end of the message,
	 * return 0 signifying end of file
	 */
	if (*msg_Ptr == 0)
		return 0;

	/*
	 * Actually put the data into the buffer
	 */
	while (length && *msg_idx) {

		/*
		 * The buffer is in the user data segment, not the kernel
		 * segment so "*" assignment won't work.  We have to use
		 * put_user which copies data from the kernel data segment to
		 * the user data segment.
		 */
		put_user(*(msg_idx++), buffer++);

		length--;
		bytes_read++;
	}

	kdbgprintln( "--- Read:-- %s", msg_Ptr);

	/*
	 * Most read functions return the number of bytes put into the buffer
	 */
	return bytes_read;
}

/*
 * Called when a process writes to dev file: echo "hi" > /dev/hello
 */
static ssize_t device_write(struct file *filp, const char *buff, size_t len, loff_t * off)
{
	int res;
	unsigned long uaddr;
	char addrstr[80];
	struct page *page;
	char *my_page_address;
	unsigned long copied = copy_from_user(addrstr, buff, sizeof(addrstr));
	if (copied != 0)
	{
	    kerrprintln("Error:Told to copy %lu, but only copied %lu\n", len, copied);
	}
	uaddr = simple_strtoul(addrstr, NULL, 0);

	down_read(&current->mm->mmap_sem);
	res = get_user_pages(current, current->mm,
	                 uaddr,
	                 1, /* Only want one page */
	                 1, /* Do want to write into it */
	                 1, /* do force */
	                 &page,
	                 NULL);
	if (res == 1)
	{

		kdbgprintln("Got page");
		/* Do something with it */
		my_page_address = kmap(page);
		kdbgprintln("Read: %s", my_page_address);
		strcpy (my_page_address, "Hello, is it me you're looking for?\n");
		kdbgprintln("Got address %p and user told me it was %lx",my_page_address, uaddr);
		kdbgprintln("Wrote: %s", my_page_address);

		kunmap(page);

		/* Clean up */
		if (!PageReserved(page))
		{
			SetPageDirty(page);
		}
		page_cache_release(page);
	}
	else
	{
	    kerrprintln("Couldn't get page :(");
	}
	up_read(&current->mm->mmap_sem);


	return len;
}
/**/
static int	chdv_mmap(struct file *filp, struct vm_area_struct *vma)
{
	static uint32_t g_cnt = 1;

	void * mem_buf_ptr;
    unsigned int pages_per_buf = ( MEM_CHUNK_SIZE >> PAGE_SHIFT );
    unsigned long pfn, vsize;

    /*make sure the buffer is allocated*/
    if(NULL == g_membuff)
    {
    	if(NULL != (g_membuff = kzalloc(mem_pool_size , GFP_KERNEL)))
    	{
    		kdbgprintln("Successfully allocated %u bytes memory pool", mem_pool_size);
    	}
    	else
    	{
			kerrprintln("Not enough memory");
			return -ENOMEM;
    	}
    }

    mem_buf_ptr = g_membuff + vma->vm_pgoff * PAGE_SIZE;
    vsize = vma->vm_end - vma->vm_start ;

    kdbgprintln("MEM_CHUNK_SIZE %lu, pages_per_buf %u, vsize %lu  vma->vm_pgoff %lu",
    		MEM_CHUNK_SIZE,
    		pages_per_buf,
    		vsize,
    		vma->vm_pgoff);
    if(vma->vm_pgoff * PAGE_SIZE + vsize > mem_pool_size)
    {
    	kerrprintln("Requested memory chunk exceeds driver memory buffer boundaries (offset %lu + size %lu) > MEM_POOL_SIZE %u",
    			vma->vm_pgoff * PAGE_SIZE, vsize, mem_pool_size);
    	return -EINVAL;
    }

    /* We allow only mapping of one whole buffer so offset must be multiple
     * of pages_per_buf and size must be equal to dma_buf_size.
     */
    if( vma->vm_pgoff % pages_per_buf )
    {
    	kerrprintln("Mapping DMA buffers is allowed only from beginning");
        return -EINVAL ;
    }

#ifdef y0el_xZx
    if( vsize != MEM_CHUNK_SIZE )
    {
    	kerrprintln("No partial mapping of DMA buffer is allowed vsize %lu != MEM_CHUNK_SIZE %u", vsize, MEM_CHUNK_SIZE);
       return -EINVAL ;
    }
#endif

    vma->vm_flags |= VM_IO | VM_PFNMAP | VM_DONTEXPAND | VM_DONTDUMP;

    pfn =  virt_to_phys(mem_buf_ptr) >> PAGE_SHIFT; //page_to_pfn(virt_to_page((unsigned char *)mem_buf_ptr))

    kdbgprintln("PFN : %lu", pfn);

    if(remap_pfn_range(vma,
    		vma->vm_start,
    		pfn,
    		vsize, vma->vm_page_prot))
    {
    	kerrprintln("Failed to remap memory");
        return -EINVAL;
    }

    memset(mem_buf_ptr, 0x00, vsize);
    /*Sealing data header & footer*/
	*((unsigned long *)mem_buf_ptr) 	 	= 0xCDFFFFFF00000000 | g_cnt;
	*((unsigned long *)mem_buf_ptr + 1) 	= 0xAB00000000000000 | g_cnt;
	*((unsigned long *)mem_buf_ptr + 7) 	= 0xAB000EaC00000000 | g_cnt;
    *(unsigned long *)((unsigned char *)mem_buf_ptr + vsize - sizeof(unsigned long)) = 0xEF00000000000000 | g_cnt;

    g_cnt++;

    kdbgprintln("Mapped & sealed'kzalloced' buffer" \
    		"\n\t\t1st  8 bytes: %lX" \
    		"\n\t\t2nd  8 bytes: %lX" \
    		"\n\t\t8th  8 bytes: %lX" \
    		"\n\t\tLast 8 bytes: %lX",
    		*((unsigned long *)mem_buf_ptr),
    		*((unsigned long *)mem_buf_ptr + 1),
    		*((unsigned long *)mem_buf_ptr + 7),
    		*(unsigned long *)((unsigned char *)mem_buf_ptr + vsize - sizeof(unsigned long)));

    return 0;
}
