/*
 ============================================================================
 Name        : chrdev_test.c
 Author      : Yoel
 Version     :
 Copyright   : Your copyright notice
 Description : character device test application
 Notes       : To compile with cuda kernels:
 cp chrdev_test.c ./chrdev_test.cu; nvcc chrdev_test.cu -L /usr/local/cuda/lib -lcudart -o chrdev_test
 ============================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <cuda_runtime_api.h>
#include <sys/mman.h>
#include <unistd.h>
#include <time.h>
#include <inttypes.h>
#include <cuda.h>

/*----------------*/
#define NANOTIME_FMT			"%"PRIu64
#define NANO_TIME_DIFF(ts, te) 	(uint64_t)(((uint64_t)((te).tv_sec - (ts).tv_sec) * ((uint64_t)1e9)) + \
							               (te).tv_nsec - (ts).tv_nsec)
#define MICRO_SLEEP 			((__useconds_t)1e6)

/*----------------*/
typedef struct
{
	void 			* address;
	cudaError_t 	stat;
}mem_info_t;
/**/
typedef struct
{
	mem_info_t 		hmai;
	mem_info_t 		cdai;
	unsigned long 	mem_size;
	unsigned long 	page_off;
}HostCudaMem;

/*----------------*/
static void     print_usage	       (char * app_name);
static void     gather_stats       (const uint64_t cptime, uint64_t* min_cptime, uint64_t* max_cptime, uint64_t* accumulate_cptime);
static void     print_data	       (unsigned long buf_id, HostCudaMem * hcm, uint16_t elem_num, uint16_t new_line_num, bool is_after_cuda);
static void     read_params	       (int argc, uint16_t * number_of_buffers, char * dev_name, char** argv);
static void 	print_stat	       (const char * stat_op_name_str, uint64_t rounds_cnt, uint64_t avg_cptime, uint64_t max_cptime, uint64_t min_cptime);
static void 	release_resources  (int fd, unsigned long buf_id, uint16_t number_of_buffers, HostCudaMem * hcm );

/*----------------*/
static unsigned long map_mem_size;

/*---------------- CUDA Kernels ----------------*/
/**/
__global__ void ckern_vec_add(uint32_t* A, uint32_t* B, uint32_t* C)
{
	int i = (blockIdx.x * blockDim.x) + threadIdx.x;

	C[i] = A[i] + B[i];
}
/**/
__global__ void ckern_mem_pad(uint64_t *a, uint64_t coef)
{
	uint64_t idx = (blockIdx.x * blockDim.x) + threadIdx.x;
	a[idx] = coef;
}
/*---------------- MAIN ----------------*/
int main(int argc, char** argv)
{
	uint16_t number_of_buffers;
	int fd;
	char dev_name[128] = {0};
	unsigned long page_off, buf_id;
	uint64_t min_cptime, max_cptime, rounds_cnt;
	uint64_t avg_cptime;
	cudaError_t cudarc;

	/*load default parameters*/
	number_of_buffers 	= 1;
	page_off 			= 0;
	map_mem_size 		= 512;
	strcpy(dev_name, "/dev/chardev");//"/dev/mvc0");


	read_params(argc, &number_of_buffers, dev_name, argv);
	map_mem_size 		*= getpagesize();
	page_off			*= getpagesize();

	HostCudaMem hcm[number_of_buffers];

	printf("App Params:\nDevice :%s" \
			"\nMap memory size  :%lu" \
			"\nPage offset      :%lu" \
			"\nNumber of buffers:%u\n",
			dev_name,
			map_mem_size,
			page_off,
			number_of_buffers);

	memset(hcm, 0x00, sizeof(hcm));

	printf("Opening %s file\n", dev_name);
	errno = 0;
	if(0 > (fd = open(dev_name, O_RDWR) ))
	{
		printf("Error: Failed to open %s device:%s\n", dev_name, strerror(errno));
	}
	else if(cudaSuccess == (cudarc = cudaSetDeviceFlags(cudaDeviceMapHost | cudaDeviceScheduleBlockingSync)))
	{
		printf("--------------------------\n");

		min_cptime = ~0x00;
		max_cptime = avg_cptime = rounds_cnt = 0;

		for(buf_id = 0; buf_id < number_of_buffers; buf_id++)
		{

			hcm[buf_id].mem_size = map_mem_size;
			hcm[buf_id].page_off = map_mem_size * buf_id;
			printf("About to map 	  buff[%04lu] %lu bytes offset %lu of %s device memory\n",
										buf_id,
										hcm[buf_id].mem_size,
										hcm[buf_id].page_off, dev_name);

			errno = 0;
			if(MAP_FAILED == (hcm[buf_id].hmai.address = mmap(NULL, hcm[buf_id].mem_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, hcm[buf_id].page_off)))
			{
				printf("Error: Failed to map %d - %s\n", errno, strerror(errno));
			}
			else
			{
				printf("About to register buff[%04lu] %lu bytes offset %lu of %s device memory to NVIDIA\n",
										buf_id,
										hcm[buf_id].mem_size,
										hcm[buf_id].page_off, dev_name);
				if ((cudaSuccess != (hcm[buf_id].hmai.stat = cudaHostRegister(hcm[buf_id].hmai.address,
																hcm[buf_id].mem_size, cudaHostRegisterIoMemory))))
				{
					printf("Error: Failed cudaHostRegister: %s, address %p\n",
								cudaGetErrorString(hcm[buf_id].hmai.stat), hcm[buf_id].hmai.address);
				}
				else
				{
					struct timespec start, end;
					int padding_val;

					print_data(buf_id, hcm, 8, 8, false);

					if (cudaSuccess != cudaHostGetDevicePointer(&(hcm[buf_id].cdai.address), hcm[buf_id].hmai.address, 0))
					{
						printf("Error: Failed cudaHostGetDevicePointer %lu bytes from buffer id %lu", hcm[buf_id].mem_size, buf_id);
					}
					else
					{
						/*Generate random value to pad the shared memory for tracking */
						clock_gettime(CLOCK_MONOTONIC, &start);
						srand(start.tv_nsec);
						padding_val = rand();

						/*Call cuda kernel*/
						ckern_mem_pad <<<1, (hcm[buf_id].mem_size) / sizeof(uint64_t)>>> ((uint64_t*)hcm[buf_id].cdai.address, (uint64_t)padding_val);

						/*start benchmark*/
						clock_gettime(CLOCK_MONOTONIC, &start);

						if (cudaSuccess != (cudarc = cudaDeviceSynchronize()))
						{
							printf("Error: Failed cudaDeviceSynchronize: %s\n", cudaGetErrorString(cudarc));
						}
						/*end benchmark*/
						clock_gettime(CLOCK_MONOTONIC, &end);

						printf("Padding Cuda random value: %X\n", padding_val);
						print_data(buf_id, hcm, 0x10, 8, true);

						gather_stats(NANO_TIME_DIFF(start, end), &min_cptime, &max_cptime, &avg_cptime);
						rounds_cnt++;
					}
				}
			}
		}
	}//End of else if(cudaSuccess == (cudarc = cudaSetDeviceFlags(cudaDeviceMapHost | cudaDeviceScheduleBlockingSync)))
	else
	{
		printf("Error: Failed cudaSetDeviceFlags: %s\n", cudaGetErrorString(cudarc));
		_exit(1);
	}

	release_resources(fd, buf_id, number_of_buffers, hcm);
	print_stat("cudaDeviceSynchronize", rounds_cnt, avg_cptime, max_cptime, min_cptime);

	return EXIT_SUCCESS;
}

/*----------------*/
/**/
static void print_usage(char * app_name)
{
	printf("Usage:   %s <device name> <number of pages> <number of buffers>\n" \
		   "Example: %s /dev/chardev 2 40\ni.e.: Need 40 buffers size of 2 pages each -> (40 x 2 x %d bytes)\n",
			app_name,
			app_name,
			getpagesize());
}
/**/
static void gather_stats(const uint64_t cptime, uint64_t* min_cptime, uint64_t* max_cptime, uint64_t* accumulate_cptime)
{
	if (*min_cptime > cptime)
	{
		*min_cptime = cptime;
	}
	if (*max_cptime < cptime)
	{
		*max_cptime = cptime;
	}
	*accumulate_cptime += cptime;
}
/**/
static void print_data(unsigned long buf_id, HostCudaMem * hcm, uint16_t elem_num, uint16_t new_line_num, bool is_after_cuda)
{
	new_line_num = (0 == new_line_num) ? 0x10 : new_line_num;

	printf("-------------%s [%lu]---------------------\n" \
		   "Start address %p",(true == is_after_cuda) ? "<-From CUDA kernel buFFer" : "->From Linux Kernel buFFer",
		   buf_id, hcm[buf_id].hmai.address);
	for (int i = 0; i < elem_num; i++)
	{
		if(0 == (i % new_line_num))
		{
			printf("\n");
		}
		printf("%04lX ",*((unsigned long *) hcm[buf_id].hmai.address + i));
	}
	/*print n - 1 buffer to make sure it is not a reflection of the current*/
	if((true == is_after_cuda) && (0 < buf_id))
	{
		printf("\n--------------<-From CUDA kernel buFFer [%lu]---------------------\n"\
			   "Start address %p", buf_id - 1, hcm[buf_id - 1].hmai.address);
		for (int i = 0; i < elem_num; i++)
		{
			if(0 == (i % new_line_num))
			{
				printf("\n");
			}
			printf("%04lX ",*((unsigned long *) hcm[buf_id - 1].hmai.address + i));
		}
	}
	printf("\n-----------------------------------------------\n");
}
/**/
static void read_params(int argc, uint16_t *number_of_buffers, char * dev_name, char** argv)
{
	switch (argc)
	{
	case 4:
		*number_of_buffers = atoi(argv[3]);
	case 3:
		map_mem_size = atoi(argv[2]);
	case 2:
		if (0 != strcmp(argv[1], "-h"))
		{
			strcpy(dev_name, argv[1]);
		}
		else
		{
			print_usage(argv[0]);
			_exit(EXIT_SUCCESS);
		}
		break;
	default:
		printf("Loading default parameters...\n");
		break;
	}
	printf("Command input:\nArgc:%d\n", argc);
	for(int i = 0; i < argc; i++)
	{
		printf("<n%d>:%s\n", i, argv[i]);
	}
}
/**/
static void print_stat(const char * stat_op_name_str, uint64_t rounds_cnt, uint64_t avg_cptime, uint64_t max_cptime, uint64_t min_cptime)
{
	if (0 != rounds_cnt)
	{
		double avg = (0 != rounds_cnt) ? (double) ((double) avg_cptime / (double) rounds_cnt) : 0.0;
		printf("%s Stats:"
				"\n\t\tMax Time:"NANOTIME_FMT" ns"
				"\n\t\tMin Time:"NANOTIME_FMT" ns"
				"\n\t\tAvg Time:%f ns\n",
				stat_op_name_str,
				max_cptime,
				(~0x00 == min_cptime) ? 0 : min_cptime, avg);
	}
}
/**/
static void release_resources(int fd, unsigned long buf_id, uint16_t number_of_buffers, HostCudaMem * hcm )
{
	if (0 <= fd)
	{
		for (buf_id = 0; buf_id < number_of_buffers; buf_id++)
		{
			if (MAP_FAILED != hcm[buf_id].hmai.address)
			{
				if (cudaSuccess == hcm[buf_id].hmai.stat)
				{
					//					printf("Unmapping CudaHost memchunk %p, offset %lu\n", hcm[buf_id].hmai.address, hcm[buf_id].page_off);
					if (cudaSuccess != (hcm[buf_id].hmai.stat = cudaHostUnregister(hcm[buf_id].hmai.address)))
					{
						printf("Error: Failed cudaHostUnregister: %s\n",
								cudaGetErrorString(hcm[buf_id].hmai.stat));
					}
				}
				//				printf("Unmapping %s file\n", dev_name);
				munmap(hcm[buf_id].hmai.address, map_mem_size);
				hcm[buf_id].hmai.address = NULL;
			}
		}
		//		printf("Closing %s file\n", dev_name);
		errno = 0;
		if (0 > close(fd))
		{
			printf("Error %d - %s\n", errno, strerror(errno));
		}
	}
}
